import { GraphQLDirective, GraphQLEnumType, GraphQLEnumValueConfigMap, GraphQLFieldConfig, GraphQLFieldConfigArgumentMap, GraphQLFieldConfigMap, GraphQLFieldResolver, GraphQLInputFieldConfig, GraphQLInputFieldConfigMap, GraphQLInputObjectType, GraphQLInputType, GraphQLInterfaceType, GraphQLNamedType, GraphQLObjectType, GraphQLOutputType, GraphQLScalarType, GraphQLUnionType, GraphQLSchema } from "graphql";
import { Metadata } from "./metadata";
import { DirectiveTypeDef, WrappedType } from "./core";
import * as Types from "./types";
/**
 * Builds all of the types, properly accounts for any using "mix".
 * Since the enum types are resolved synchronously, these need to guard for
 * circular references at this step, while fields will guard for it during lazy evaluation.
 */
export declare class SchemaBuilder {
    protected metadata: Metadata;
    protected config: Types.BuilderConfig;
    /**
     * Used to check for circular references.
     */
    protected buildingTypes: Set<any>;
    /**
     * The "final type" map contains all types as they are built.
     */
    protected finalTypeMap: Record<string, GraphQLNamedType>;
    /**
     * The "defined type" map keeps track of all of the types that were
     * defined directly as `GraphQL*Type` objects, so we don't accidentally
     * overwrite any.
     */
    protected definedTypeMap: Record<string, GraphQLNamedType>;
    /**
     * The "pending type" map keeps track of all types that were defined w/
     * GraphQL Nexus and haven't been processed into concrete types yet.
     */
    protected pendingTypeMap: Record<string, Types.NamedTypeDef>;
    protected pendingDirectiveMap: Record<string, DirectiveTypeDef<any>>;
    protected directiveMap: Record<string, GraphQLDirective>;
    protected nullability: Types.NullabilityConfig;
    constructor(metadata: Metadata, config: Types.BuilderConfig);
    getConfig(): Types.BuilderConfig;
    addType(typeDef: Types.NamedTypeDef | GraphQLNamedType): void;
    addDirective(directiveDef: DirectiveTypeDef<any> | GraphQLDirective): void;
    getFinalTypeMap(): Types.BuildTypes<any, any>;
    directiveType(config: Types.DirectiveTypeConfig): GraphQLDirective;
    inputObjectType(config: Types.InputTypeConfig): GraphQLInputObjectType;
    objectType(config: Types.ObjectTypeConfig): GraphQLObjectType;
    interfaceType(config: Types.InterfaceTypeConfig): GraphQLInterfaceType;
    enumType(config: Types.EnumTypeConfig): GraphQLEnumType;
    unionType(config: Types.UnionTypeConfig): GraphQLUnionType;
    protected missingType(typeName: string): GraphQLNamedType;
    protected buildEnumMembers(config: Types.EnumTypeConfig): GraphQLEnumValueConfigMap;
    protected buildUnionMembers(config: Types.UnionTypeConfig): GraphQLObjectType[];
    protected buildObjectFields(typeConfig: Types.ObjectTypeConfig | Types.InterfaceTypeConfig): GraphQLFieldConfigMap<any, any>;
    protected buildInputObjectFields(typeConfig: Types.InputTypeConfig): GraphQLInputFieldConfigMap;
    protected buildObjectField(fieldConfig: Types.OutputFieldConfig, typeConfig: Types.ObjectTypeConfig | Types.InterfaceTypeConfig): GraphQLFieldConfig<any, any>;
    protected buildInputObjectField(field: Types.InputFieldConfig, typeConfig: Types.InputTypeConfig): GraphQLInputFieldConfig;
    protected buildArgs(args: Types.OutputFieldArgs, typeConfig: Types.InputTypeConfig): GraphQLFieldConfigArgumentMap;
    protected decorateInputType(type: GraphQLInputType, fieldConfig: Types.InputFieldConfig, typeConfig: Types.InputTypeConfig): GraphQLInputType;
    protected decorateOutputType(type: GraphQLOutputType, fieldConfig: Types.FieldConfig, typeConfig: Types.ObjectTypeConfig | Types.InterfaceTypeConfig): GraphQLOutputType;
    protected decorateArgType(type: GraphQLInputType, argOpts: Types.ArgDefinition & {
        name: string;
    }, typeConfig: Types.InputTypeConfig): GraphQLInputType;
    /**
     * Adds the null / list configuration to the type.
     */
    protected decorateType(type: GraphQLOutputType, fieldConfig: Types.Omit<Types.FieldConfig, "type" | "default">, typeConfig: Types.ObjectTypeConfig | Types.InterfaceTypeConfig, isInput: false): GraphQLOutputType;
    protected decorateType(type: GraphQLInputType, fieldConfig: Types.Omit<Types.FieldConfig, "type" | "default">, typeConfig: Types.InputTypeConfig, isInput: true): GraphQLInputType;
    protected getInterface(name: string): GraphQLInterfaceType;
    protected getEnum(name: string): GraphQLEnumType;
    protected getUnion(name: string): GraphQLUnionType;
    protected getInputType(name: string): GraphQLScalarType | GraphQLEnumType | GraphQLInputObjectType;
    protected getOutputType(name: string): GraphQLObjectType | GraphQLInterfaceType | GraphQLUnionType | GraphQLScalarType | GraphQLEnumType;
    protected getObjectType(name: string): GraphQLObjectType;
    protected getOrBuildType(name: string): GraphQLNamedType;
    protected getResolver(fieldOptions: Types.OutputFieldConfig, typeConfig: Types.ObjectTypeConfig | Types.InterfaceTypeConfig): GraphQLFieldResolver<any, any, {
        [argName: string]: any;
    }>;
}
/**
 * Builds the types, normalizing the "types" passed into the schema for a
 * better developer experience. This is primarily useful for testing
 * type generation
 */
export declare function buildTypes<TypeMapDefs extends Record<string, GraphQLNamedType> = any, DirectiveDefs extends Record<string, GraphQLDirective> = any>(types: any, config?: Types.BuilderConfig, SchemaBuilderClass?: typeof SchemaBuilder, MetadataClass?: typeof Metadata): Types.BuildTypes<TypeMapDefs, DirectiveDefs>;
/**
 * Builds the schema, returning both the schema and metadata.
 */
export declare function makeSchemaWithMetadata(options: Types.SchemaConfig, SchemaBuilderClass?: typeof SchemaBuilder, MetadataClass?: typeof Metadata): {
    metadata: Metadata;
    schema: GraphQLSchema;
};
/**
 * Defines the GraphQL schema, by combining the GraphQL types defined
 * by the GraphQL Nexus layer or any manually defined GraphQLType objects.
 *
 * Requires at least one type be named "Query", which will be used as the
 * root query type.
 */
export declare function makeSchema(options: Types.SchemaConfig): GraphQLSchema;
export declare function isWrappedTypeDef(obj: any): obj is WrappedType;
export declare function isNamedTypeDef(obj: any): obj is Types.NamedTypeDef;
