"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var tslib_1 = require("tslib");
var graphql_1 = require("graphql");
var metadata_1 = require("./metadata");
var core_1 = require("./core");
var Types = tslib_1.__importStar(require("./types"));
var utils_1 = require("./utils");
var util_1 = require("util");
var isPromise = function (val) {
    return Boolean(val && typeof val.then === "function");
};
var NULL_DEFAULTS = {
    output: false,
    outputList: false,
    outputListItem: false,
    input: true,
    inputList: true,
    inputListItem: false,
};
var SCALARS = {
    String: graphql_1.GraphQLString,
    Int: graphql_1.GraphQLInt,
    Float: graphql_1.GraphQLFloat,
    ID: graphql_1.GraphQLID,
    Boolean: graphql_1.GraphQLBoolean,
};
/**
 * Builds all of the types, properly accounts for any using "mix".
 * Since the enum types are resolved synchronously, these need to guard for
 * circular references at this step, while fields will guard for it during lazy evaluation.
 */
var SchemaBuilder = /** @class */ (function () {
    function SchemaBuilder(metadata, config) {
        this.metadata = metadata;
        this.config = config;
        /**
         * Used to check for circular references.
         */
        this.buildingTypes = new Set();
        /**
         * The "final type" map contains all types as they are built.
         */
        this.finalTypeMap = {};
        /**
         * The "defined type" map keeps track of all of the types that were
         * defined directly as `GraphQL*Type` objects, so we don't accidentally
         * overwrite any.
         */
        this.definedTypeMap = {};
        /**
         * The "pending type" map keeps track of all types that were defined w/
         * GraphQL Nexus and haven't been processed into concrete types yet.
         */
        this.pendingTypeMap = {};
        this.pendingDirectiveMap = {};
        this.directiveMap = {};
        this.nullability = {};
        this.nullability = config.nullability || {};
    }
    SchemaBuilder.prototype.getConfig = function () {
        return this.config;
    };
    SchemaBuilder.prototype.addType = function (typeDef) {
        var existingType = this.finalTypeMap[typeDef.name] || this.pendingTypeMap[typeDef.name];
        if (existingType) {
            // Allow importing the same exact type more than once.
            if (existingType === typeDef) {
                return;
            }
            throw extendError(typeDef.name);
        }
        if (graphql_1.isNamedType(typeDef)) {
            this.metadata.addExternalType(typeDef);
            this.finalTypeMap[typeDef.name] = typeDef;
            this.definedTypeMap[typeDef.name] = typeDef;
        }
        else {
            this.pendingTypeMap[typeDef.name] = typeDef;
        }
    };
    SchemaBuilder.prototype.addDirective = function (directiveDef) {
        if (graphql_1.isDirective(directiveDef)) {
            this.directiveMap[directiveDef.name] = directiveDef;
        }
        else {
            this.pendingDirectiveMap[directiveDef.name] = directiveDef;
        }
    };
    SchemaBuilder.prototype.getFinalTypeMap = function () {
        var _this = this;
        Object.keys(this.pendingTypeMap).forEach(function (key) {
            // If we've already constructed the type by this point,
            // via circular dependency resolution don't worry about building it.
            if (_this.finalTypeMap[key]) {
                return;
            }
            if (_this.definedTypeMap[key]) {
                throw extendError(key);
            }
            _this.finalTypeMap[key] = _this.getOrBuildType(key);
            _this.buildingTypes.clear();
        });
        Object.keys(this.pendingDirectiveMap).forEach(function (key) { });
        return {
            typeMap: this.finalTypeMap,
            metadata: this.metadata,
            directiveMap: this.directiveMap,
        };
    };
    SchemaBuilder.prototype.directiveType = function (config) {
        return new graphql_1.GraphQLDirective({
            name: config.name,
            locations: [],
        });
    };
    SchemaBuilder.prototype.inputObjectType = function (config) {
        var _this = this;
        return new graphql_1.GraphQLInputObjectType({
            name: config.name,
            fields: function () { return _this.buildInputObjectFields(config); },
            description: config.description,
        });
    };
    SchemaBuilder.prototype.objectType = function (config) {
        var _this = this;
        this.metadata.addObjectType(config);
        return new graphql_1.GraphQLObjectType({
            name: config.name,
            interfaces: function () { return config.interfaces.map(function (i) { return _this.getInterface(i); }); },
            description: config.description,
            fields: function () {
                var interfaceFields = {};
                var allInterfaces = config.interfaces.map(function (i) {
                    return _this.getInterface(i);
                });
                allInterfaces.forEach(function (i) {
                    var iFields = i.getFields();
                    // We need to take the interface fields and reconstruct them
                    // this actually simplifies things becuase if we've modified
                    // the field at all it needs to happen here.
                    Object.keys(iFields).forEach(function (iFieldName) {
                        var _a = iFields[iFieldName], isDeprecated = _a.isDeprecated, args = _a.args, rest = tslib_1.__rest(_a, ["isDeprecated", "args"]);
                        interfaceFields[iFieldName] = tslib_1.__assign({}, rest, { args: args.reduce(function (result, arg) {
                                var name = arg.name, argRest = tslib_1.__rest(arg, ["name"]);
                                result[name] = argRest;
                                return result;
                            }, {}) });
                    });
                });
                return tslib_1.__assign({}, interfaceFields, _this.buildObjectFields(config));
            },
        });
    };
    SchemaBuilder.prototype.interfaceType = function (config) {
        var _this = this;
        var name = config.name, resolveType = config.resolveType, description = config.description;
        return new graphql_1.GraphQLInterfaceType({
            name: name,
            fields: function () { return _this.buildObjectFields(config); },
            resolveType: resolveType,
            description: description,
        });
    };
    SchemaBuilder.prototype.enumType = function (config) {
        return new graphql_1.GraphQLEnumType({
            name: config.name,
            values: this.buildEnumMembers(config),
        });
    };
    SchemaBuilder.prototype.unionType = function (config) {
        var _this = this;
        return new graphql_1.GraphQLUnionType({
            name: config.name,
            types: function () { return _this.buildUnionMembers(config); },
            resolveType: config.resolveType,
        });
    };
    SchemaBuilder.prototype.missingType = function (typeName) {
        var suggestions = utils_1.suggestionList(typeName, Object.keys(this.buildingTypes).concat(Object.keys(this.finalTypeMap)));
        var suggestionsString = "";
        if (suggestions.length > 0) {
            suggestionsString = " or mean " + suggestions.join(", ");
        }
        throw new Error("Missing type " + typeName + ", did you forget to import a type" + suggestionsString + "?");
    };
    SchemaBuilder.prototype.buildEnumMembers = function (config) {
        var _this = this;
        var values = {};
        config.members.forEach(function (member) {
            switch (member.item) {
                case Types.NodeType.ENUM_MEMBER:
                    values[member.info.name] = {
                        value: member.info.value,
                        description: member.info.description,
                    };
                    break;
                case Types.NodeType.MIX:
                    var _a = member.mixOptions, pick_1 = _a.pick, omit_1 = _a.omit, typeName = member.typeName;
                    var enumToMix = _this.getEnum(typeName);
                    enumToMix.getValues().forEach(function (val) {
                        if (pick_1 && pick_1.indexOf(val.name) === -1) {
                            return;
                        }
                        if (omit_1 && omit_1.indexOf(val.name) !== -1) {
                            return;
                        }
                        values[val.name] = {
                            description: val.description,
                            deprecationReason: val.deprecationReason,
                            value: val.value,
                        };
                    });
            }
        });
        if (!Object.keys(values).length) {
            throw new Error("GraphQL Nexus: Enum " + config.name + " must have at least one member");
        }
        return values;
    };
    SchemaBuilder.prototype.buildUnionMembers = function (config) {
        var _this = this;
        var unionMembers = [];
        config.members.forEach(function (member) {
            switch (member.item) {
                case Types.NodeType.UNION_MEMBER:
                    unionMembers.push(_this.getObjectType(member.typeName));
                    break;
                case Types.NodeType.MIX:
                    var _a = member.mixOptions, pick_2 = _a.pick, omit_2 = _a.omit, typeName = member.typeName;
                    var unionToMix = _this.getUnion(typeName);
                    unionToMix.getTypes().forEach(function (type) {
                        if (pick_2 && pick_2.indexOf(type.name) === -1) {
                            return;
                        }
                        if (omit_2 && omit_2.indexOf(type.name) !== -1) {
                            return;
                        }
                        unionMembers.push(type);
                    });
                    break;
            }
        });
        if (!Object.keys(unionMembers).length) {
            throw new Error("GraphQL Nexus: Union " + config.name + " must have at least one member type");
        }
        return unionMembers;
    };
    SchemaBuilder.prototype.buildObjectFields = function (typeConfig) {
        var _this = this;
        var fieldMap = {};
        typeConfig.fields.forEach(function (field) {
            switch (field.item) {
                case Types.NodeType.MIX:
                    throw new Error("TODO");
                    break;
                case Types.NodeType.FIELD:
                    fieldMap[field.config.name] = _this.buildObjectField(field.config, typeConfig);
                    break;
            }
        });
        return fieldMap;
    };
    SchemaBuilder.prototype.buildInputObjectFields = function (typeConfig) {
        var _this = this;
        var fieldMap = {};
        typeConfig.fields.forEach(function (field) {
            switch (field.item) {
                case Types.NodeType.MIX:
                    throw new Error("TODO");
                    break;
                case Types.NodeType.FIELD:
                    fieldMap[field.config.name] = _this.buildInputObjectField(field.config, typeConfig);
                    break;
            }
        });
        return fieldMap;
    };
    SchemaBuilder.prototype.buildObjectField = function (fieldConfig, typeConfig) {
        this.metadata.addField(typeConfig.name, fieldConfig);
        return {
            type: this.decorateOutputType(this.getOutputType(fieldConfig.type), fieldConfig, typeConfig),
            resolve: this.getResolver(fieldConfig, typeConfig),
            description: fieldConfig.description,
            args: this.buildArgs(fieldConfig.args || {}, typeConfig),
        };
    };
    SchemaBuilder.prototype.buildInputObjectField = function (field, typeConfig) {
        return {
            type: this.decorateInputType(this.getInputType(field.type), field, typeConfig),
        };
    };
    SchemaBuilder.prototype.buildArgs = function (args, typeConfig) {
        var _this = this;
        var allArgs = {};
        Object.keys(args).forEach(function (argName) {
            var argDef = args[argName];
            allArgs[argName] = {
                type: _this.decorateArgType(_this.getInputType(argDef.type), tslib_1.__assign({}, argDef, { name: argName }), typeConfig),
                description: argDef.description,
            };
        });
        return allArgs;
    };
    SchemaBuilder.prototype.decorateInputType = function (type, fieldConfig, typeConfig) {
        var _required = fieldConfig.required, requiredListItem = fieldConfig.requiredListItem, rest = tslib_1.__rest(fieldConfig, ["required", "requiredListItem"]);
        var newOpts = rest;
        if (typeof _required !== "undefined") {
            newOpts.nullable = !_required;
        }
        if (typeof requiredListItem !== "undefined") {
            if (rest.list) {
                newOpts.listItemNullable = !requiredListItem;
            }
        }
        return this.decorateType(type, newOpts, typeConfig, true);
    };
    SchemaBuilder.prototype.decorateOutputType = function (type, fieldConfig, typeConfig) {
        return this.decorateType(type, fieldConfig, typeConfig, false);
    };
    SchemaBuilder.prototype.decorateArgType = function (type, argOpts, typeConfig) {
        var _required = argOpts.required, requiredListItem = argOpts.requiredListItem, rest = tslib_1.__rest(argOpts, ["required", "requiredListItem"]);
        var newOpts = rest;
        if (typeof _required !== "undefined") {
            newOpts.nullable = !_required;
        }
        if (typeof requiredListItem !== "undefined") {
            if (rest.list) {
                newOpts.listItemNullable = !requiredListItem;
            }
        }
        return this.decorateType(type, newOpts, typeConfig, true);
    };
    SchemaBuilder.prototype.decorateType = function (type, fieldConfig, typeConfig, isInput) {
        var finalType = type;
        var nullConfig = tslib_1.__assign({}, NULL_DEFAULTS, this.nullability, typeConfig.nullability);
        var list = fieldConfig.list, nullable = fieldConfig.nullable, listDepth = fieldConfig.listDepth, listItemNullable = fieldConfig.listItemNullable;
        var isNullable = typeof nullable !== "undefined"
            ? nullable
            : list
                ? isInput
                    ? nullConfig.inputList
                    : nullConfig.outputList
                : isInput
                    ? nullConfig.input
                    : nullConfig.output;
        if (list) {
            var depth = listDepth || 1;
            var nullableItem = typeof listItemNullable !== "undefined"
                ? listItemNullable
                : isInput
                    ? nullConfig.inputListItem
                    : nullConfig.outputListItem;
            if (Array.isArray(nullableItem) && nullableItem.length !== depth) {
                throw new Error("Incorrect listItemNullable array length for " + typeConfig.name + fieldConfig.name + ", expected " + depth + " saw " + nullableItem.length);
            }
            for (var i = 0; i < depth; i++) {
                var isNull = Array.isArray(nullableItem)
                    ? nullableItem[i]
                    : nullableItem;
                if (!isNull) {
                    finalType = graphql_1.GraphQLNonNull(finalType);
                }
                finalType = graphql_1.GraphQLList(finalType);
            }
        }
        else if (typeof listItemNullable !== "undefined") {
            console.log("listItemNullable should only be set with list: true, this option is ignored");
        }
        if (!isNullable) {
            return graphql_1.GraphQLNonNull(finalType);
        }
        return finalType;
    };
    SchemaBuilder.prototype.getInterface = function (name) {
        var type = this.getOrBuildType(name);
        if (!graphql_1.isInterfaceType(type)) {
            throw new Error("Expected " + name + " to be an interfaceType, saw " + type.constructor.name);
        }
        return type;
    };
    SchemaBuilder.prototype.getEnum = function (name) {
        var type = this.getOrBuildType(name);
        if (!graphql_1.isEnumType(type)) {
            throw new Error("Expected " + name + " to be an enumType, saw " + type.constructor.name);
        }
        return type;
    };
    SchemaBuilder.prototype.getUnion = function (name) {
        var type = this.getOrBuildType(name);
        if (!graphql_1.isUnionType(type)) {
            throw new Error("Expected " + name + " to be a unionType, saw " + type.constructor.name);
        }
        return type;
    };
    SchemaBuilder.prototype.getInputType = function (name) {
        var type = this.getOrBuildType(name);
        if (!graphql_1.isInputType(type)) {
            throw new Error("Expected " + name + " to be a valid input type, saw " + type.constructor.name);
        }
        return type;
    };
    SchemaBuilder.prototype.getOutputType = function (name) {
        var type = this.getOrBuildType(name);
        if (!graphql_1.isOutputType(type)) {
            throw new Error("Expected " + name + " to be a valid output type, saw " + type.constructor.name);
        }
        return type;
    };
    SchemaBuilder.prototype.getObjectType = function (name) {
        var type = this.getOrBuildType(name);
        if (!graphql_1.isObjectType(type)) {
            throw new Error("Expected " + name + " to be a objectType, saw " + type.constructor.name);
        }
        return type;
    };
    SchemaBuilder.prototype.getOrBuildType = function (name) {
        if (SCALARS[name]) {
            return SCALARS[name];
        }
        if (this.finalTypeMap[name]) {
            return this.finalTypeMap[name];
        }
        if (this.buildingTypes.has(name)) {
            throw new Error("GraphQL Nexus: Circular dependency detected, while building types " + Array.from(this.buildingTypes));
        }
        var pendingType = this.pendingTypeMap[name];
        if (pendingType) {
            this.buildingTypes.add(name);
            return pendingType.buildType(this);
        }
        return this.missingType(name);
    };
    SchemaBuilder.prototype.getResolver = function (fieldOptions, typeConfig) {
        var resolver = typeConfig.defaultResolver || graphql_1.defaultFieldResolver;
        if (fieldOptions.resolve) {
            resolver = fieldOptions.resolve;
        }
        if (typeof fieldOptions.default !== "undefined") {
            resolver = withDefaultValue(resolver, fieldOptions.default);
        }
        return resolver;
    };
    return SchemaBuilder;
}());
exports.SchemaBuilder = SchemaBuilder;
function withDefaultValue(resolver, defaultValue) {
    return function (root, args, ctx, info) {
        var result = resolver(root, args, ctx, info);
        if (typeof result === "undefined" || result === null) {
            return typeof defaultValue === "function" ? defaultValue() : defaultValue;
        }
        if (isPromise(result)) {
            return result.then(function (val) {
                if (typeof val === "undefined" || val === null) {
                    return typeof defaultValue === "function"
                        ? defaultValue()
                        : defaultValue;
                }
                return val;
            });
        }
        return result;
    };
}
function extendError(name) {
    return new Error(name + " was already defined as a type, check the docs for extending");
}
/**
 * Builds the types, normalizing the "types" passed into the schema for a
 * better developer experience. This is primarily useful for testing
 * type generation
 */
function buildTypes(types, config, SchemaBuilderClass, MetadataClass) {
    if (config === void 0) { config = { outputs: false }; }
    if (SchemaBuilderClass === void 0) { SchemaBuilderClass = SchemaBuilder; }
    if (MetadataClass === void 0) { MetadataClass = metadata_1.Metadata; }
    var metadata = new MetadataClass(config);
    var builder = new SchemaBuilderClass(metadata, config);
    addTypes(builder, types);
    return builder.getFinalTypeMap();
}
exports.buildTypes = buildTypes;
function addTypes(builder, types) {
    if (!types) {
        return;
    }
    if (isWrappedTypeDef(types)) {
        types = types.type;
        if (typeof types === "function") {
            addTypes(builder, types(builder));
            return;
        }
    }
    if (isNamedTypeDef(types) || graphql_1.isNamedType(types)) {
        builder.addType(types);
    }
    else if (types instanceof core_1.DirectiveTypeDef || graphql_1.isDirective(types)) {
        builder.addDirective(types);
    }
    else if (Array.isArray(types)) {
        types.forEach(function (typeDef) { return addTypes(builder, typeDef); });
    }
    else if (util_1.isObject(types)) {
        Object.keys(types).forEach(function (key) { return addTypes(builder, types[key]); });
    }
}
/**
 * Builds the schema, returning both the schema and metadata.
 */
function makeSchemaWithMetadata(options, SchemaBuilderClass, MetadataClass) {
    if (SchemaBuilderClass === void 0) { SchemaBuilderClass = SchemaBuilder; }
    if (MetadataClass === void 0) { MetadataClass = metadata_1.Metadata; }
    var _a = buildTypes(options.types, options, SchemaBuilderClass, MetadataClass), typeMap = _a.typeMap, directiveMap = _a.directiveMap, metadata = _a.metadata;
    var Query = typeMap.Query, Mutation = typeMap.Mutation, Subscription = typeMap.Subscription;
    if (!Query) {
        console.warn("GraphQL Nexus: You should define a root `Query` type for your schema");
        Query = new graphql_1.GraphQLObjectType({
            name: "Query",
            fields: {
                ok: {
                    type: graphql_1.GraphQLNonNull(graphql_1.GraphQLBoolean),
                    resolve: function () { return true; },
                },
            },
        });
    }
    if (!graphql_1.isObjectType(Query)) {
        throw new Error("Expected Query to be a objectType, saw " + Query.constructor.name);
    }
    if (Mutation && !graphql_1.isObjectType(Mutation)) {
        throw new Error("Expected Mutation to be a objectType, saw " + Mutation.constructor.name);
    }
    if (Subscription && !graphql_1.isObjectType(Subscription)) {
        throw new Error("Expected Subscription to be a objectType, saw " + Subscription.constructor.name);
    }
    var schema = new graphql_1.GraphQLSchema({
        query: Query,
        mutation: Mutation,
        subscription: Subscription,
        directives: graphql_1.specifiedDirectives.concat(utils_1.objValues(directiveMap)),
        types: utils_1.objValues(typeMap),
    });
    metadata.finishConstruction();
    return { schema: schema, metadata: metadata };
}
exports.makeSchemaWithMetadata = makeSchemaWithMetadata;
/**
 * Defines the GraphQL schema, by combining the GraphQL types defined
 * by the GraphQL Nexus layer or any manually defined GraphQLType objects.
 *
 * Requires at least one type be named "Query", which will be used as the
 * root query type.
 */
function makeSchema(options) {
    var _a = makeSchemaWithMetadata(options), schema = _a.schema, metadata = _a.metadata;
    // Only in development envs do we want to worry about regenerating the
    // schema definition and/or generated types.
    var _b = options.shouldGenerateArtifacts, shouldGenerateArtifacts = _b === void 0 ? process.env.NODE_ENV !== "production" : _b;
    if (shouldGenerateArtifacts) {
        // Generating in the next tick allows us to use the schema
        // in the optional thunk for the typegen config
        metadata.generateArtifacts(schema);
    }
    return schema;
}
exports.makeSchema = makeSchema;
function isWrappedTypeDef(obj) {
    return obj instanceof core_1.WrappedType;
}
exports.isWrappedTypeDef = isWrappedTypeDef;
function isNamedTypeDef(obj) {
    return (obj instanceof core_1.ObjectTypeDef ||
        obj instanceof core_1.InputObjectTypeDef ||
        obj instanceof core_1.EnumTypeDef ||
        obj instanceof core_1.UnionTypeDef ||
        obj instanceof core_1.InterfaceTypeDef);
}
exports.isNamedTypeDef = isNamedTypeDef;
//# sourceMappingURL=builder.js.map