import * as Types from "./types";
import { GraphQLFieldResolver, GraphQLInputObjectType, GraphQLInterfaceType, GraphQLUnionType, GraphQLEnumType, GraphQLIsTypeOfFn, DirectiveLocationEnum, GraphQLObjectType, GraphQLDirective, GraphQLScalarType } from "graphql";
import { SchemaBuilder, isNamedTypeDef } from "./builder";
import { Metadata } from "./metadata";
export { SDLConverter } from "./sdlConverter";
export { Types };
export { SchemaBuilder, isNamedTypeDef, Metadata };
export { typegenAutoConfig } from "./autoConfig";
declare global {
    interface GraphQLNexusGen {
    }
}
/**
 * Provided to the [objectType](#objectType) function, this
 */
export declare class ObjectTypeDef<GenTypes = GraphQLNexusGen, TypeName extends string = any> {
    readonly name: string;
    /**
     * All metadata about the object type
     */
    protected typeConfig: Types.ObjectTypeConfig;
    constructor(name: string);
    /**
     * Mixes in an existing field definition or object type
     * with the current type.
     */
    mix(typeName: string, options?: Types.MixOpts<any>): void;
    /**
     * Add an ID field type to the object schema.
     */
    id<FieldName extends string>(name: FieldName, ...opts: Types.ConditionalOutputFieldOpts<GenTypes, TypeName, FieldName>): void;
    /**
     * Add an Int field type to the object schema.
     */
    int<FieldName extends string>(name: FieldName, ...opts: Types.ConditionalOutputFieldOpts<GenTypes, TypeName, FieldName>): void;
    /**
     * Add a Float field type to the object schema.
     */
    float<FieldName extends string>(name: FieldName, ...opts: Types.ConditionalOutputFieldOpts<GenTypes, TypeName, FieldName>): void;
    /**
     * Add a String field type to the object schema.
     */
    string<FieldName extends string>(name: FieldName, ...opts: Types.ConditionalOutputFieldOpts<GenTypes, TypeName, FieldName>): void;
    /**
     * Add a Boolean field type to the object schema.
     */
    boolean<FieldName extends string>(name: FieldName, ...opts: Types.ConditionalOutputFieldOpts<GenTypes, TypeName, FieldName>): void;
    /**
     * Adds a new field to the object type
     */
    field<FieldName extends string>(name: FieldName, type: Types.AllOutputTypes<GenTypes> | Types.BaseScalars, ...opts: Types.ConditionalOutputFieldOpts<GenTypes, TypeName, FieldName>): void;
    /**
     * Declare that an object type implements a particular interface,
     * by providing the name of the interface
     */
    implements(...interfaceName: Types.AllInterfaces<GenTypes>[]): void;
    /**
     * Adds a description to the `GraphQLObjectType`
     *
     * Descriptions will be output as type annotations in the generated SDL
     */
    description(description: string): void;
    /**
     * Adds an "isTypeOf" check to the object type.
     */
    isTypeOf(fn: (value: GraphQLIsTypeOfFn<any, any>) => boolean): void;
    /**
     * Used to modify a field already defined on an interface or mixed-in
     * from another type.
     *
     * At this point the type will not change, but the resolver,
     * default, property, or description fields can.
     */
    modify<FieldName extends Types.ObjectTypeFields<GenTypes, TypeName>>(field: FieldName, options: Types.ModifyFieldOpts<GenTypes, TypeName, FieldName>): void;
    /**
     * Supply the default field resolver for all members of this type
     */
    defaultResolver(resolverFn: GraphQLFieldResolver<Types.RootValue<GenTypes, TypeName>, Types.ContextValue<GenTypes>>): void;
    /**
     * Adds a directive directly to the object definition
     *
     * > Should be used rarely, typically only for interpretation by other schema consumers.
     */
    directive(name: string, args?: Record<string, any>): void;
    /**
     * Configures the nullability for the type, check the
     * documentation's "Getting Started" section to learn
     * more about GraphQL Nexus's assumptions and configuration
     * on nullability.
     *
     * @param nullability
     */
    nullability(nullability: Types.NullabilityConfig): void;
    /**
     * @internal
     */
    buildType(builder: SchemaBuilder): GraphQLObjectType;
}
/**
 * Backing type for an enum member.
 */
export declare class EnumTypeDef<GenTypes = GraphQLNexusGen> {
    readonly name: string;
    protected typeConfig: Types.EnumTypeConfig;
    constructor(name: string);
    mix<EnumName extends Types.EnumNames<GenTypes>>(typeName: EnumName, options?: Types.MixOpts<Types.EnumMembers<GenTypes, EnumName>>): void;
    member(name: string, config?: Types.EnumMemberConfig): void;
    /**
     * Sets the members of the enum
     */
    members(info: Array<Types.EnumMemberInfo | string>): void;
    /**
     * Adds a description to the `GraphQLEnumType`
     *
     * Descriptions will be output as type annotations in the generated SDL
     */
    description(description: string): void;
    /**
     * Should be used very rarely, adds a directive directly to the
     * enum definition - for interpretation by other schema consumers.
     */
    directive(name: string, args?: Record<string, any>): void;
    /**
     * @internal
     */
    buildType(builder: SchemaBuilder): GraphQLEnumType;
}
/**
 * Configure the `GraphQLUnionType` definition
 */
export declare class UnionTypeDef<GenTypes = GraphQLNexusGen, TypeName extends string = any> {
    readonly name: string;
    protected typeConfig: Types.UnionTypeConfig;
    constructor(name: string);
    /**
     * Take an existing union and base the type off of that, using the `omit`
     * option to exclude members of the other union.
     *
     * > Note: Circular dependencies between unions are not allowed and will
     * trigger an error at build-time.
     */
    mix<UnionTypeName extends string>(type: UnionTypeName, options?: Types.MixOmitOpts<any>): void;
    /**
     * Add one or more members to the GraphQLUnion. Any types provided should be valid
     * object types available to the schema.
     */
    members(...types: Array<Types.ObjectNames<GenTypes>>): void;
    /**
     * Define a type resolver function for the union type. The Resolver should
     * return the type name of the union member that should be fulfilled.
     *
     * > Providing this is highly recommended. If one is not provided, the
     * default implementation will call `isTypeOf` on each implementing Object type.
     *
     * @see https://github.com/graphql/graphql-js/issues/876#issuecomment-304398882
     */
    resolveType(typeResolver: Types.TypeResolver<GenTypes, TypeName>): void;
    /**
     * Should be used very rarely, adds a directive directly to the
     * union definition - for interpretation by other schema consumers.
     */
    directive(name: string, args?: Record<string, any>): void;
    /**
     * @internal
     */
    buildType(builder: SchemaBuilder): GraphQLUnionType;
}
export declare class InterfaceTypeDef<GenTypes = GraphQLNexusGen, TypeName extends string = any> {
    readonly name: string;
    /**
     * Metadata about the object type
     */
    protected typeConfig: Types.InterfaceTypeConfig;
    constructor(name: string);
    /**
     * Mixes in an existing field definition or object type
     * with the current type.
     */
    mix(typeName: string, options?: Types.MixOpts<any>): void;
    /**
     * Add an ID field type to the object schema.
     */
    id<FieldName extends string>(name: FieldName, ...opts: Types.ConditionalOutputFieldOpts<GenTypes, TypeName, FieldName>): void;
    /**
     * Add an Int field type to the object schema.
     */
    int<FieldName extends string>(name: FieldName, ...opts: Types.ConditionalOutputFieldOpts<GenTypes, TypeName, FieldName>): void;
    /**
     * Add a Float field type to the object schema.
     */
    float<FieldName extends string>(name: FieldName, ...opts: Types.ConditionalOutputFieldOpts<GenTypes, TypeName, FieldName>): void;
    /**
     * Add a String field type to the object schema.
     */
    string<FieldName extends string>(name: FieldName, ...opts: Types.ConditionalOutputFieldOpts<GenTypes, TypeName, FieldName>): void;
    /**
     * Add a Boolean field type to the object schema.
     */
    boolean<FieldName extends string>(name: FieldName, ...opts: Types.ConditionalOutputFieldOpts<GenTypes, TypeName, FieldName>): void;
    /**
     * Adds a new field to the object type
     */
    field<FieldName extends string>(name: FieldName, type: Types.AllOutputTypes<GenTypes> | Types.BaseScalars, ...opts: Types.ConditionalOutputFieldOpts<GenTypes, TypeName, FieldName>): void;
    /**
     * Adds a description to the `GraphQLInterfaceType`
     *
     * Descriptions will be output as type annotations in the generated SDL
     */
    description(description: string): void;
    /**
     * Optionally provide a custom type resolver function. If one is not provided,
     * the default implementation will call `isTypeOf` on each implementing
     * Object type.
     */
    resolveType(typeResolver: Types.TypeResolver<GenTypes, TypeName>): void;
    /**
     * Should be used very rarely, adds a directive directly to the
     * interface definition - for interpretation by other schema consumers.
     */
    directive(name: string, args?: Record<string, any>): void;
    /**
     * @internal
     */
    buildType(builder: SchemaBuilder): GraphQLInterfaceType;
}
export declare class InputObjectTypeDef<GenTypes = GraphQLNexusGen, TypeName extends string = any> {
    readonly name: string;
    protected typeConfig: Types.InputTypeConfig;
    constructor(name: string);
    /**
     * Add an ID field type to the object schema.
     */
    id(name: string, options?: Types.InputFieldOpts<GenTypes, "ID">): void;
    /**
     * Add an Int field type to the object schema.
     */
    int(name: string, options?: Types.InputFieldOpts<GenTypes, "Int">): void;
    /**
     * Add a Float field type to the object schema.
     */
    float(name: string, options?: Types.InputFieldOpts<GenTypes, "Float">): void;
    /**
     * Add a String field type to the object schema.
     */
    string(name: string, options?: Types.InputFieldOpts<GenTypes, "String">): void;
    /**
     * Add a Boolean field type to the object schema.
     */
    boolean(name: string, options?: Types.InputFieldOpts<GenTypes, "Boolean">): void;
    /**
     * Adds a new field to the input object type
     */
    field<FieldName extends string>(name: FieldName, type: Types.AllInputTypes<GenTypes> | Types.BaseScalars, options?: Types.InputFieldOpts<GenTypes, TypeName>): void;
    /**
     * Adds a description to the `GraphQLInputObjectType`
     *
     * Descriptions will be output as type annotations in the generated SDL
     */
    description(description: string): void;
    /**
     * Should be used very rarely, adds a directive directly to the
     * input object definition - for interpretation by other schema consumers.
     */
    directive(name: string, args?: Record<string, any>): void;
    /**
     * Configures the nullability for the type
     *
     * @see nullability
     */
    nullability(nullability: Types.NullabilityConfig): void;
    /**
     * @internal
     */
    buildType(builder: SchemaBuilder): GraphQLInputObjectType;
}
export declare class DirectiveTypeDef<GenTypes = GraphQLNexusGen> {
    readonly name: string;
    protected typeConfig: Types.DirectiveTypeConfig;
    constructor(name: string);
    description(description: string): void;
    locations(...location: DirectiveLocationEnum[]): void;
    /**
     * Add an ID field type to the object schema.
     */
    id(name: string, options?: Types.InputFieldOpts<GenTypes, "ID">): void;
    /**
     * Add an Int field type to the object schema.
     */
    int(name: string, options?: Types.InputFieldOpts<GenTypes, "Int">): void;
    /**
     * Add a Float field type to the object schema.
     */
    float(name: string, options?: Types.InputFieldOpts<GenTypes, "Float">): void;
    /**
     * Add a String field type to the object schema.
     */
    string(name: string, options?: Types.InputFieldOpts<GenTypes, "String">): void;
    /**
     * Add a Boolean field type to the object schema.
     */
    boolean(name: string, options?: Types.InputFieldOpts<GenTypes, "Boolean">): void;
    /**
     * Adds a new field to the input object type
     */
    field<TypeName extends Types.AllInputTypes<GenTypes> | Types.BaseScalars>(name: string, type: TypeName, options?: Types.InputFieldOpts<GenTypes, TypeName>): void;
    /**
     * @internal
     */
    buildType(builder: SchemaBuilder): GraphQLDirective;
}
/**
 * The `WrappedType` exists to signify that the value returned from
 * the type construction APIs should not be used externally outside of the
 * builder function. It also is useful if you need the SchemaBuilder, in that
 * it can take a function which is lazy-evaluated to build the type.
 */
export declare class WrappedType {
    readonly type: Types.NamedTypeDef | DirectiveTypeDef | GraphQLScalarType | ((schema: SchemaBuilder) => WrappedType);
    constructor(type: Types.NamedTypeDef | DirectiveTypeDef | GraphQLScalarType | ((schema: SchemaBuilder) => WrappedType));
}
