"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var tslib_1 = require("tslib");
var Types = tslib_1.__importStar(require("./types"));
exports.Types = Types;
var utils_1 = require("./utils");
var builder_1 = require("./builder");
exports.SchemaBuilder = builder_1.SchemaBuilder;
exports.isNamedTypeDef = builder_1.isNamedTypeDef;
var metadata_1 = require("./metadata");
exports.Metadata = metadata_1.Metadata;
var sdlConverter_1 = require("./sdlConverter");
exports.SDLConverter = sdlConverter_1.SDLConverter;
// Keeping this in core since it shouldn't be needed directly
var autoConfig_1 = require("./autoConfig");
exports.typegenAutoConfig = autoConfig_1.typegenAutoConfig;
/**
 * Provided to the [objectType](#objectType) function, this
 */
var ObjectTypeDef = /** @class */ (function () {
    function ObjectTypeDef(name) {
        this.name = name;
        this.typeConfig = {
            name: name,
            fields: [],
            interfaces: [],
            directives: [],
            fieldModifications: {},
        };
    }
    /**
     * Mixes in an existing field definition or object type
     * with the current type.
     */
    ObjectTypeDef.prototype.mix = function (typeName, options) {
        this.typeConfig.fields.push({
            item: Types.NodeType.MIX,
            typeName: typeName,
            mixOptions: options || {},
        });
    };
    /**
     * Add an ID field type to the object schema.
     */
    ObjectTypeDef.prototype.id = function (name) {
        var opts = [];
        for (var _i = 1; _i < arguments.length; _i++) {
            opts[_i - 1] = arguments[_i];
        }
        this.field.apply(this, [name, "ID"].concat(opts));
    };
    /**
     * Add an Int field type to the object schema.
     */
    ObjectTypeDef.prototype.int = function (name) {
        var opts = [];
        for (var _i = 1; _i < arguments.length; _i++) {
            opts[_i - 1] = arguments[_i];
        }
        this.field.apply(this, [name, "Int"].concat(opts));
    };
    /**
     * Add a Float field type to the object schema.
     */
    ObjectTypeDef.prototype.float = function (name) {
        var opts = [];
        for (var _i = 1; _i < arguments.length; _i++) {
            opts[_i - 1] = arguments[_i];
        }
        this.field.apply(this, [name, "Float"].concat(opts));
    };
    /**
     * Add a String field type to the object schema.
     */
    ObjectTypeDef.prototype.string = function (name) {
        var opts = [];
        for (var _i = 1; _i < arguments.length; _i++) {
            opts[_i - 1] = arguments[_i];
        }
        this.field.apply(this, [name, "String"].concat(opts));
    };
    /**
     * Add a Boolean field type to the object schema.
     */
    ObjectTypeDef.prototype.boolean = function (name) {
        var opts = [];
        for (var _i = 1; _i < arguments.length; _i++) {
            opts[_i - 1] = arguments[_i];
        }
        this.field.apply(this, [name, "Boolean"].concat(opts));
    };
    /**
     * Adds a new field to the object type
     */
    ObjectTypeDef.prototype.field = function (name, type) {
        var opts = [];
        for (var _i = 2; _i < arguments.length; _i++) {
            opts[_i - 2] = arguments[_i];
        }
        var options = {};
        if (typeof opts[0] === "function") {
            options.resolve = opts[0];
        }
        else {
            options = tslib_1.__assign({}, opts[0]);
        }
        this.typeConfig.fields.push({
            item: Types.NodeType.FIELD,
            config: tslib_1.__assign({ name: name,
                type: type }, options),
        });
    };
    /**
     * Declare that an object type implements a particular interface,
     * by providing the name of the interface
     */
    ObjectTypeDef.prototype.implements = function () {
        var interfaceName = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            interfaceName[_i] = arguments[_i];
        }
        var _a;
        (_a = this.typeConfig.interfaces).push.apply(_a, interfaceName);
    };
    /**
     * Adds a description to the `GraphQLObjectType`
     *
     * Descriptions will be output as type annotations in the generated SDL
     */
    ObjectTypeDef.prototype.description = function (description) {
        this.typeConfig.description = utils_1.dedent(description);
    };
    /**
     * Adds an "isTypeOf" check to the object type.
     */
    ObjectTypeDef.prototype.isTypeOf = function (fn) {
        this.typeConfig.isTypeOf = fn;
    };
    /**
     * Used to modify a field already defined on an interface or mixed-in
     * from another type.
     *
     * At this point the type will not change, but the resolver,
     * default, property, or description fields can.
     */
    ObjectTypeDef.prototype.modify = function (field, options) {
        this.typeConfig.fieldModifications[field] = options;
    };
    /**
     * Supply the default field resolver for all members of this type
     */
    ObjectTypeDef.prototype.defaultResolver = function (resolverFn) {
        this.typeConfig.defaultResolver = resolverFn;
    };
    /**
     * Adds a directive directly to the object definition
     *
     * > Should be used rarely, typically only for interpretation by other schema consumers.
     */
    ObjectTypeDef.prototype.directive = function (name, args) {
        this.typeConfig.directives.push({
            name: name,
            args: args || {},
        });
    };
    /**
     * Configures the nullability for the type, check the
     * documentation's "Getting Started" section to learn
     * more about GraphQL Nexus's assumptions and configuration
     * on nullability.
     *
     * @param nullability
     */
    ObjectTypeDef.prototype.nullability = function (nullability) {
        if (this.typeConfig.nullability) {
            console.warn("nullability has already been set for type " + this.typeConfig.name + ", the previous value will be replaced");
        }
        this.typeConfig.nullability = nullability;
    };
    /**
     * @internal
     */
    ObjectTypeDef.prototype.buildType = function (builder) {
        return builder.objectType(this.typeConfig);
    };
    return ObjectTypeDef;
}());
exports.ObjectTypeDef = ObjectTypeDef;
/**
 * Backing type for an enum member.
 */
var EnumTypeDef = /** @class */ (function () {
    function EnumTypeDef(name) {
        this.name = name;
        this.typeConfig = {
            name: name,
            members: [],
            directives: [],
        };
    }
    EnumTypeDef.prototype.mix = function (typeName, options) {
        this.typeConfig.members.push({
            item: Types.NodeType.MIX,
            typeName: typeName,
            mixOptions: options || {},
        });
    };
    EnumTypeDef.prototype.member = function (name, config) {
        this.typeConfig.members.push({
            item: Types.NodeType.ENUM_MEMBER,
            info: tslib_1.__assign({ name: name, value: name }, config),
        });
    };
    /**
     * Sets the members of the enum
     */
    EnumTypeDef.prototype.members = function (info) {
        var _this = this;
        info.forEach(function (member) {
            if (typeof member === "string") {
                return _this.typeConfig.members.push({
                    item: Types.NodeType.ENUM_MEMBER,
                    info: { name: member, value: member },
                });
            }
            _this.typeConfig.members.push({
                item: Types.NodeType.ENUM_MEMBER,
                info: member,
            });
        });
    };
    /**
     * Adds a description to the `GraphQLEnumType`
     *
     * Descriptions will be output as type annotations in the generated SDL
     */
    EnumTypeDef.prototype.description = function (description) {
        this.typeConfig.description = utils_1.dedent(description);
    };
    /**
     * Should be used very rarely, adds a directive directly to the
     * enum definition - for interpretation by other schema consumers.
     */
    EnumTypeDef.prototype.directive = function (name, args) {
        this.typeConfig.directives.push({
            name: name,
            args: args || {},
        });
    };
    /**
     * @internal
     */
    EnumTypeDef.prototype.buildType = function (builder) {
        return builder.enumType(this.typeConfig);
    };
    return EnumTypeDef;
}());
exports.EnumTypeDef = EnumTypeDef;
/**
 * Configure the `GraphQLUnionType` definition
 */
var UnionTypeDef = /** @class */ (function () {
    function UnionTypeDef(name) {
        this.name = name;
        this.typeConfig = {
            name: name,
            members: [],
            directives: [],
        };
    }
    /**
     * Take an existing union and base the type off of that, using the `omit`
     * option to exclude members of the other union.
     *
     * > Note: Circular dependencies between unions are not allowed and will
     * trigger an error at build-time.
     */
    UnionTypeDef.prototype.mix = function (type, options) {
        this.typeConfig.members.push({
            item: Types.NodeType.MIX,
            typeName: type,
            mixOptions: options || {},
        });
    };
    /**
     * Add one or more members to the GraphQLUnion. Any types provided should be valid
     * object types available to the schema.
     */
    UnionTypeDef.prototype.members = function () {
        var _this = this;
        var types = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            types[_i] = arguments[_i];
        }
        types.forEach(function (typeName) {
            _this.typeConfig.members.push({
                item: Types.NodeType.UNION_MEMBER,
                typeName: typeName,
            });
        });
    };
    /**
     * Define a type resolver function for the union type. The Resolver should
     * return the type name of the union member that should be fulfilled.
     *
     * > Providing this is highly recommended. If one is not provided, the
     * default implementation will call `isTypeOf` on each implementing Object type.
     *
     * @see https://github.com/graphql/graphql-js/issues/876#issuecomment-304398882
     */
    UnionTypeDef.prototype.resolveType = function (typeResolver) {
        this.typeConfig.resolveType = typeResolver;
    };
    /**
     * Should be used very rarely, adds a directive directly to the
     * union definition - for interpretation by other schema consumers.
     */
    UnionTypeDef.prototype.directive = function (name, args) {
        this.typeConfig.directives.push({
            name: name,
            args: args || {},
        });
    };
    /**
     * @internal
     */
    UnionTypeDef.prototype.buildType = function (builder) {
        return builder.unionType(this.typeConfig);
    };
    return UnionTypeDef;
}());
exports.UnionTypeDef = UnionTypeDef;
var InterfaceTypeDef = /** @class */ (function () {
    function InterfaceTypeDef(name) {
        this.name = name;
        this.typeConfig = {
            name: name,
            fields: [],
            directives: [],
        };
    }
    /**
     * Mixes in an existing field definition or object type
     * with the current type.
     */
    InterfaceTypeDef.prototype.mix = function (typeName, options) {
        this.typeConfig.fields.push({
            item: Types.NodeType.MIX,
            typeName: typeName,
            mixOptions: options || {},
        });
    };
    /**
     * Add an ID field type to the object schema.
     */
    InterfaceTypeDef.prototype.id = function (name) {
        var opts = [];
        for (var _i = 1; _i < arguments.length; _i++) {
            opts[_i - 1] = arguments[_i];
        }
        this.field.apply(this, [name, "ID"].concat(opts));
    };
    /**
     * Add an Int field type to the object schema.
     */
    InterfaceTypeDef.prototype.int = function (name) {
        var opts = [];
        for (var _i = 1; _i < arguments.length; _i++) {
            opts[_i - 1] = arguments[_i];
        }
        this.field.apply(this, [name, "Int"].concat(opts));
    };
    /**
     * Add a Float field type to the object schema.
     */
    InterfaceTypeDef.prototype.float = function (name) {
        var opts = [];
        for (var _i = 1; _i < arguments.length; _i++) {
            opts[_i - 1] = arguments[_i];
        }
        this.field.apply(this, [name, "Float"].concat(opts));
    };
    /**
     * Add a String field type to the object schema.
     */
    InterfaceTypeDef.prototype.string = function (name) {
        var opts = [];
        for (var _i = 1; _i < arguments.length; _i++) {
            opts[_i - 1] = arguments[_i];
        }
        this.field.apply(this, [name, "String"].concat(opts));
    };
    /**
     * Add a Boolean field type to the object schema.
     */
    InterfaceTypeDef.prototype.boolean = function (name) {
        var opts = [];
        for (var _i = 1; _i < arguments.length; _i++) {
            opts[_i - 1] = arguments[_i];
        }
        this.field.apply(this, [name, "Boolean"].concat(opts));
    };
    /**
     * Adds a new field to the object type
     */
    InterfaceTypeDef.prototype.field = function (name, type) {
        var opts = [];
        for (var _i = 2; _i < arguments.length; _i++) {
            opts[_i - 2] = arguments[_i];
        }
        var options = {};
        if (typeof opts[0] === "function") {
            options.resolve = opts[0];
        }
        else {
            options = tslib_1.__assign({}, opts[0]);
        }
        this.typeConfig.fields.push({
            item: Types.NodeType.FIELD,
            config: tslib_1.__assign({ name: name,
                type: type }, options),
        });
    };
    /**
     * Adds a description to the `GraphQLInterfaceType`
     *
     * Descriptions will be output as type annotations in the generated SDL
     */
    InterfaceTypeDef.prototype.description = function (description) {
        this.typeConfig.description = utils_1.dedent(description);
    };
    /**
     * Optionally provide a custom type resolver function. If one is not provided,
     * the default implementation will call `isTypeOf` on each implementing
     * Object type.
     */
    InterfaceTypeDef.prototype.resolveType = function (typeResolver) {
        this.typeConfig.resolveType = typeResolver;
    };
    /**
     * Should be used very rarely, adds a directive directly to the
     * interface definition - for interpretation by other schema consumers.
     */
    InterfaceTypeDef.prototype.directive = function (name, args) {
        this.typeConfig.directives.push({
            name: name,
            args: args || {},
        });
    };
    /**
     * @internal
     */
    InterfaceTypeDef.prototype.buildType = function (builder) {
        return builder.interfaceType(this.typeConfig);
    };
    return InterfaceTypeDef;
}());
exports.InterfaceTypeDef = InterfaceTypeDef;
var InputObjectTypeDef = /** @class */ (function () {
    function InputObjectTypeDef(name) {
        this.name = name;
        this.typeConfig = {
            name: name,
            fields: [],
            directives: [],
        };
    }
    /**
     * Add an ID field type to the object schema.
     */
    InputObjectTypeDef.prototype.id = function (name, options) {
        this.field(name, "ID", options);
    };
    /**
     * Add an Int field type to the object schema.
     */
    InputObjectTypeDef.prototype.int = function (name, options) {
        this.field(name, "Int", options);
    };
    /**
     * Add a Float field type to the object schema.
     */
    InputObjectTypeDef.prototype.float = function (name, options) {
        this.field(name, "Float", options);
    };
    /**
     * Add a String field type to the object schema.
     */
    InputObjectTypeDef.prototype.string = function (name, options) {
        this.field(name, "String", options);
    };
    /**
     * Add a Boolean field type to the object schema.
     */
    InputObjectTypeDef.prototype.boolean = function (name, options) {
        this.field(name, "Boolean", options);
    };
    /**
     * Adds a new field to the input object type
     */
    InputObjectTypeDef.prototype.field = function (name, type, options) {
        this.typeConfig.fields.push({
            item: Types.NodeType.FIELD,
            config: tslib_1.__assign({ name: name,
                type: type }, options),
        });
    };
    /**
     * Adds a description to the `GraphQLInputObjectType`
     *
     * Descriptions will be output as type annotations in the generated SDL
     */
    InputObjectTypeDef.prototype.description = function (description) {
        this.typeConfig.description = utils_1.dedent(description);
    };
    /**
     * Should be used very rarely, adds a directive directly to the
     * input object definition - for interpretation by other schema consumers.
     */
    InputObjectTypeDef.prototype.directive = function (name, args) {
        this.typeConfig.directives.push({
            name: name,
            args: args || {},
        });
    };
    /**
     * Configures the nullability for the type
     *
     * @see nullability
     */
    InputObjectTypeDef.prototype.nullability = function (nullability) {
        if (this.typeConfig.nullability) {
            console.warn("nullability has already been set for type " + this.typeConfig.name + ", the previous value will be replaced");
        }
        this.typeConfig.nullability = nullability;
    };
    /**
     * @internal
     */
    InputObjectTypeDef.prototype.buildType = function (builder) {
        return builder.inputObjectType(this.typeConfig);
    };
    return InputObjectTypeDef;
}());
exports.InputObjectTypeDef = InputObjectTypeDef;
var DirectiveTypeDef = /** @class */ (function () {
    function DirectiveTypeDef(name) {
        this.name = name;
        this.typeConfig = {
            name: name,
            locations: [],
            directiveArgs: [],
        };
    }
    DirectiveTypeDef.prototype.description = function (description) {
        this.typeConfig.description = utils_1.dedent(description);
    };
    DirectiveTypeDef.prototype.locations = function () {
        var location = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            location[_i] = arguments[_i];
        }
        var _a;
        (_a = this.typeConfig.locations).push.apply(_a, location);
    };
    /**
     * Add an ID field type to the object schema.
     */
    DirectiveTypeDef.prototype.id = function (name, options) {
        this.field(name, "ID", options);
    };
    /**
     * Add an Int field type to the object schema.
     */
    DirectiveTypeDef.prototype.int = function (name, options) {
        this.field(name, "Int", options);
    };
    /**
     * Add a Float field type to the object schema.
     */
    DirectiveTypeDef.prototype.float = function (name, options) {
        this.field(name, "Float", options);
    };
    /**
     * Add a String field type to the object schema.
     */
    DirectiveTypeDef.prototype.string = function (name, options) {
        this.field(name, "String", options);
    };
    /**
     * Add a Boolean field type to the object schema.
     */
    DirectiveTypeDef.prototype.boolean = function (name, options) {
        this.field(name, "Boolean", options);
    };
    /**
     * Adds a new field to the input object type
     */
    DirectiveTypeDef.prototype.field = function (name, type, options) {
        this.typeConfig.directiveArgs.push(tslib_1.__assign({ name: name,
            type: type }, options));
    };
    /**
     * @internal
     */
    DirectiveTypeDef.prototype.buildType = function (builder) {
        return builder.directiveType(this.typeConfig);
    };
    return DirectiveTypeDef;
}());
exports.DirectiveTypeDef = DirectiveTypeDef;
/**
 * The `WrappedType` exists to signify that the value returned from
 * the type construction APIs should not be used externally outside of the
 * builder function. It also is useful if you need the SchemaBuilder, in that
 * it can take a function which is lazy-evaluated to build the type.
 */
var WrappedType = /** @class */ (function () {
    function WrappedType(type) {
        this.type = type;
    }
    return WrappedType;
}());
exports.WrappedType = WrappedType;
//# sourceMappingURL=core.js.map