"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var tslib_1 = require("tslib");
var graphql_1 = require("graphql");
var core_1 = require("./core");
var utils_1 = require("./utils");
/**
 * The most basic components of a GraphQL schema are object types, which just represent
 * a kind of object you can fetch from your service, and what fields it has.
 *
 * ```
 * const User = objectType('User', (t) => {
 *   t.int('id', { description: 'Id of the user' })
 *   t.string('fullName', { description: 'Full name of the user' })
 *   t.field('status', 'StatusEnum');
 *   t.field('posts', 'Post', {
 *     list: true,
 *     resolve(root, args, ctx) {
 *       return ctx.getUser(root.id).posts()
 *     }
 *   })
 * });
 *
 * const Post = objectType('Post', (t) => {
 *   t.int('id')
 *   t.string('title')
 * })
 *
 * const StatusEnum = enumType('StatusEnum', {
 *   ACTIVE: 1,
 *   DISABLED: 2
 * });
 * ```
 *
 * @see https://graphql.github.io/learn/schema/#object-types-and-fields
 */
function objectType(name, fn) {
    var factory = new core_1.ObjectTypeDef(graphql_1.assertValidName(name));
    fn(factory);
    return new core_1.WrappedType(factory);
}
exports.objectType = objectType;
/**
 * Like many type systems, GraphQL supports interfaces. An Interface is an
 * abstract type that includes a certain set of fields that a type must
 * include to implement the interface.
 *
 * In GraphQL Nexus, you do not need to redefine the interface fields on the
 * implementing object types, instead you may use `.implements(interfaceName)`
 * and all of the interface fields will be added to the type.
 *
 * ```
 * const Node = interfaceType('Node', t => {
 *   t.id('id', { description: 'GUID for a resource' });
 * });
 *
 * const User = objectType('User', t => {
 *   t.implements('Node');
 * });
 * ```
 *
 * @see https://graphql.github.io/learn/schema/#interfaces
 */
function interfaceType(name, fn) {
    var factory = new core_1.InterfaceTypeDef(graphql_1.assertValidName(name));
    fn(factory);
    return new core_1.WrappedType(factory);
}
exports.interfaceType = interfaceType;
/**
 * Union types are very similar to interfaces, but they don't get to specify
 * any common fields between the types.
 *
 * As a function, where other unions can be mixed in:
 *
 * ```
 * const CombinedResult = unionType('CombinedResult', t => {
 *   t.mix('SearchResult')
 *   t.members('AnotherType', 'YetAnotherType')
 *   t.resolveType(item => item.name)
 * })
 * ```
 *
 * @see https://graphql.org/learn/schema/#union-types
 */
function unionType(name, fn) {
    var factory = new core_1.UnionTypeDef(graphql_1.assertValidName(name));
    fn(factory);
    return new core_1.WrappedType(factory);
}
exports.unionType = unionType;
/**
 * A Enum is a special GraphQL type that represents a set of symbolic names (members)
 * bound to unique, constant values. There are three ways to create a GraphQLEnumType
 * with enumType:
 *
 * As an array of enum values:
 *
 * ```
 * const Episode = enumType('Episode', ['NEWHOPE', 'EMPIRE', 'JEDI'])
 * ```
 *
 * As an object, with a mapping of enum values to internal values:
 *
 * ```
 * const Episode = enumType('Episode', {
 *   NEWHOPE: 4,
 *   EMPIRE: 5,
 *   JEDI: 6
 * });
 * ```
 *
 * As a function, where other enums can be mixed in:
 *
 * ```
 * const Episode = enumType('Episode', (t) => {
 *   t.mix('OneThroughThree')
 *   t.mix('FourThroughSix')
 *   t.mix('SevenThroughNine')
 *   t.members(['OTHER'])
 *   t.description('All Movies in the Skywalker saga, or OTHER')
 * })
 * ```
 *
 * @see https://graphql.github.io/learn/schema/#enumeration-types
 */
function enumType(name, fn) {
    var factory = new core_1.EnumTypeDef(graphql_1.assertValidName(name));
    if (typeof fn === "function") {
        fn(factory);
    }
    else {
        factory.members(utils_1.enumShorthandMembers(fn));
    }
    return new core_1.WrappedType(factory);
}
exports.enumType = enumType;
/**
 * Defines a complex object which can be passed as an input value.
 *
 * Unlike object types, input types do not have arguments, and they do not
 * have resolvers, backing types, etc.
 *
 * @see https://graphql.org/learn/schema/#input-types
 */
function inputObjectType(name, fn) {
    var factory = new core_1.InputObjectTypeDef(graphql_1.assertValidName(name));
    fn(factory);
    return new core_1.WrappedType(factory);
}
exports.inputObjectType = inputObjectType;
/**
 * A GraphQL object type has a name and fields, but at some point those fields have
 * to resolve to some concrete data. That's where the scalar types come in:
 * they represent the leaves of the query.
 *
 * ```js
 * const DateScalar = scalarType("Date", {
 *   description: "Date custom scalar type",
 *   parseValue(value) {
 *     return new Date(value);
 *   },
 *   serialize(value) {
 *     return value.getTime();
 *   },
 *   parseLiteral(ast) {
 *     if (ast.kind === Kind.INT) {
 *       return new Date(ast.value);
 *     }
 *     return null;
 *   }
 * });
 * ```
 *
 * @see https://graphql.github.io/learn/schema/#scalar-types
 */
function scalarType(name, options) {
    return new core_1.WrappedType(new graphql_1.GraphQLScalarType(tslib_1.__assign({ name: graphql_1.assertValidName(name) }, options)));
}
exports.scalarType = scalarType;
/**
 * Defines an argument that can be used in any object or interface type
 *
 * Takes the GraphQL type name and any options.
 *
 * The value returned from this argument can be used multiple times in any valid `args` object value
 *
 * @see https://graphql.github.io/learn/schema/#arguments
 */
function arg(type, options) {
    // This isn't wrapped for now because it's not a named type, it's really
    // just an object that can be reused in multiple locations.
    return tslib_1.__assign({ type: type }, options);
}
exports.arg = arg;
/**
 * Alias for `arg("String", options)`
 */
function stringArg(options) {
    return arg("String", options);
}
exports.stringArg = stringArg;
/**
 * Alias for `arg("Int", options)`
 */
function intArg(options) {
    return arg("Int", options);
}
exports.intArg = intArg;
/**
 * Alias for `arg("Float", options)`
 */
function floatArg(options) {
    return arg("Float", options);
}
exports.floatArg = floatArg;
/**
 * Alias for `arg("ID", options)`
 */
function idArg(options) {
    return arg("ID", options);
}
exports.idArg = idArg;
/**
 * Alias for `arg("Boolean", options)`
 */
function booleanArg(options) {
    return arg("Boolean", options);
}
exports.booleanArg = booleanArg;
/**
 * Defines a directive that can be used by the schema.
 *
 * > Note: Directives should be rarely used, as they only function for external
 * > consumers of the schema.
 */
function directiveType(name, config) {
    var directive = new core_1.DirectiveTypeDef(graphql_1.assertValidName(name));
    if (typeof config === "function") {
        config(directive);
    }
    else {
        directive.locations.apply(directive, config.locations);
    }
    return new core_1.WrappedType(directive);
}
exports.directiveType = directiveType;
//# sourceMappingURL=definitions.js.map