"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var tslib_1 = require("tslib");
var graphql_1 = require("graphql");
var path_1 = tslib_1.__importDefault(require("path"));
var lang_1 = require("./lang");
var utils_1 = require("./utils");
var typegen_1 = require("./typegen");
var autoConfig_1 = require("./autoConfig");
var common_1 = require("./common");
var prettierFormat_1 = require("./prettierFormat");
/**
 * Passed into the SchemaBuilder, this keeps track of any necessary
 * field / type metadata we need to be aware of when building the
 * generated types and/or SDL artifact, including but not limited to:
 *
 * - Directive usage
 * - The TS type backing the GraphQL type
 * - Type default resolver
 * - Whether the field has a resolver
 * - The property name of the field
 * - If the field has a "default" value
 */
var Metadata = /** @class */ (function () {
    function Metadata(config) {
        this.config = config;
        this.completed = false;
        this.objectMeta = {};
        this.interfaceMeta = {};
        this.objectFieldMeta = {};
        this.typeImports = [];
        this.typegenFile = "";
        if (config.outputs !== false && config.shouldGenerateArtifacts !== false) {
            if (config.outputs.typegen) {
                this.typegenFile = utils_1.assertAbsolutePath(config.outputs.typegen, "outputs.typegen");
            }
        }
    }
    Metadata.prototype.finishConstruction = function () {
        this.completed = true;
    };
    /**
     * Ensure the type doesn't conflict with an existing type, prefixing
     * with a _ if it does
     */
    Metadata.prototype.safeTypeName = function (schema, typeName) {
        if (!schema.getType(typeName)) {
            return typeName;
        }
        return "_" + typeName;
    };
    /**
     * Check for the field's existence in an object type.
     */
    Metadata.prototype.hasField = function (schema, typeName, fieldName) {
        var type = schema.getType(typeName);
        if (!type) {
            throw new Error(".hasField should only be called on known type names");
        }
        if (graphql_1.isInputObjectType(type) ||
            graphql_1.isObjectType(type) ||
            graphql_1.isInterfaceType(type)) {
            var fields = type.getFields();
            return Boolean(fields[fieldName]);
        }
        throw new Error(".hasField should only be used with GraphQL types with fields, " + type);
    };
    // Predicates:
    Metadata.prototype.isExternalType = function (typeName) {
        return Boolean(this.objectMeta[typeName]);
    };
    Metadata.prototype.hasResolver = function (typeName, fieldName) {
        if (this.isFieldModified(typeName, fieldName)) {
        }
        return Boolean(this.objectFieldMeta[typeName] &&
            this.objectFieldMeta[typeName][fieldName] &&
            this.objectFieldMeta[typeName][fieldName].resolve);
    };
    Metadata.prototype.hasDefaultResolver = function (typeName) {
        return Boolean(this.objectMeta[typeName] && this.objectMeta[typeName].defaultResolver);
    };
    Metadata.prototype.hasDefaultValue = function (type, fieldName) {
        return Boolean(this.objectFieldMeta[type.name] &&
            this.objectFieldMeta[type.name][fieldName] &&
            this.objectFieldMeta[type.name][fieldName].default);
    };
    Metadata.prototype.isFieldModified = function (typeName, fieldName) {
        return Boolean(this.objectMeta[typeName] &&
            this.objectMeta[typeName].fieldModifications[fieldName]);
    };
    Metadata.prototype.isInterfaceField = function (type, fieldName) {
        return Boolean(type.getInterfaces().forEach(function (i) { return i.getFields()[fieldName]; }));
    };
    // Schema construction helpers:
    Metadata.prototype.addScalar = function (config) {
        this.checkMutable();
    };
    Metadata.prototype.addInterfaceType = function (config) {
        this.checkMutable();
        this.interfaceMeta[config.name] = config;
    };
    Metadata.prototype.addObjectType = function (config) {
        this.checkMutable();
        this.objectMeta[config.name] = config;
    };
    Metadata.prototype.addExternalType = function (type) {
        this.checkMutable();
    };
    Metadata.prototype.addField = function (typeName, field) {
        this.checkMutable();
        this.objectFieldMeta[typeName] = this.objectFieldMeta[typeName] || {};
        this.objectFieldMeta[typeName][field.name] = field;
    };
    /**
     * Generates the artifacts of the build based on what we
     * know about the schema and how it was defined.
     */
    Metadata.prototype.generateArtifacts = function (schema) {
        var _this = this;
        var sortedSchema = this.sortSchema(schema);
        if (this.config.outputs) {
            if (this.config.outputs.schema) {
                this.writeFile("schema", this.generateSchemaFile(sortedSchema), this.config.outputs.schema);
            }
            var typegen_2 = this.config.outputs.typegen;
            if (typegen_2) {
                this.generateTypesFile(sortedSchema).then(function (value) {
                    return _this.writeFile("types", value, typegen_2);
                });
            }
        }
    };
    Metadata.prototype.sortSchema = function (schema) {
        var sortedSchema = schema;
        if (typeof graphql_1.lexicographicSortSchema !== "undefined") {
            sortedSchema = graphql_1.lexicographicSortSchema(schema);
        }
        return sortedSchema;
    };
    Metadata.prototype.writeFile = function (type, output, filePath) {
        if (typeof filePath !== "string" || !path_1.default.isAbsolute(filePath)) {
            return Promise.reject(new Error("Expected an absolute path to output the GraphQL Nexus " + type + ", saw " + filePath));
        }
        var fs = require("fs");
        var util = require("util");
        var _a = [
            util.promisify(fs.readFile),
            util.promisify(fs.writeFile),
        ], readFile = _a[0], writeFile = _a[1];
        var formatTypegen = null;
        if (typeof this.config.formatTypegen === "function") {
            formatTypegen = this.config.formatTypegen;
        }
        else if (this.config.prettierConfig) {
            formatTypegen = prettierFormat_1.prettierFormat(this.config.prettierConfig);
        }
        var content = Promise.resolve(typeof formatTypegen === "function" ? formatTypegen(output, type) : output);
        return Promise.all([
            content,
            readFile(filePath, "utf8").catch(function () { return ""; }),
        ]).then(function (_a) {
            var toSave = _a[0], existing = _a[1];
            if (toSave !== existing) {
                return writeFile(filePath, toSave);
            }
        });
    };
    /**
     * Generates the schema, adding any directives as necessary
     */
    Metadata.prototype.generateSchemaFile = function (schema) {
        var printedSchema = graphql_1.printSchema(schema);
        /**
         * If there are directives defined to be used on the types,
         * we need to add these manually to the AST. Directives shouldn't
         * be too common, since we're defining the schema programatically
         * rather than by hand.
         */
        if (Object.keys({}).length > 0) {
            printedSchema = graphql_1.printSchema(graphql_1.visit(graphql_1.parse(printedSchema), {
            // TODO: Add directives
            }));
        }
        return [lang_1.SDL_HEADER, printedSchema].join("\n\n");
    };
    /**
     * Generates the type definitions
     */
    Metadata.prototype.generateTypesFile = function (schema) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            return tslib_1.__generator(this, function (_a) {
                return [2 /*return*/, typegen_1.buildTypeDefinitions(schema, this)];
            });
        });
    };
    Metadata.prototype.getTypegenInfo = function (schema) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            return tslib_1.__generator(this, function (_a) {
                if (this.config.typegenConfig) {
                    if (this.config.typegenAutoConfig) {
                        console.warn("Only one of typegenConfig and typegenAutoConfig should be specified, ignoring typegenConfig");
                    }
                    return [2 /*return*/, this.config.typegenConfig(schema, this.typegenFile)];
                }
                else if (this.config.typegenAutoConfig) {
                    return [2 /*return*/, autoConfig_1.typegenAutoConfig(this.config.typegenAutoConfig)(schema, this.typegenFile)];
                }
                return [2 /*return*/, {
                        headers: [lang_1.TYPEGEN_HEADER],
                        imports: [],
                        contextType: "unknown",
                        backingTypeMap: tslib_1.__assign({}, common_1.SCALAR_TYPES),
                    }];
            });
        });
    };
    Metadata.prototype.checkMutable = function () {
        if (this.completed) {
            throw new Error("Metadata cannot be modified after the schema is built");
        }
    };
    return Metadata;
}());
exports.Metadata = Metadata;
//# sourceMappingURL=metadata.js.map